;;;; Trivial Captcha Generator

;;;; Works best with captchas of around 6 letters. Use pnmtopng from
;;;; the netpbm package to make a png of the output, like so:
;;;;    (make-captcha "wibble")
;;;;    pnmtopng /tmp/captcha.pnm > /tmp/captcha.png

;;;; Copyright 2005 Tim Daly Jr.
(defpackage :trivial-captcha
  (:use :cl)
  (:export #:make-captcha))

(in-package :trivial-captcha)

(defparameter *letter-database*
  (make-pathname :name "letters" :type "pnm"
                 :defaults (truename (asdf:system-definition-pathname :trivial-captcha)))
  "This is where the letters are taken from.")

(defun make-captcha (&optional (letters "foobar") (output-path "/tmp/captcha.pnm"))
  ;; read in the letters
  (let ((letter-images (with-open-file (s *letter-database*)
			 (read-pgm s))))
    ;; copy some letters
    (let* ((letter-width 24)
	   (letter-height 48)
	   (captcha (make-array (list (* (length letters) letter-width)
				      letter-height))))
	(loop for l across letters
	   for letter-offset = (* letter-width
				  (position l "abcdefghijklmnopqrstuvwxyz"))
	     for i upfrom 0 by letter-width
	     do
	     (dotimes (di letter-width)
	       (dotimes (j letter-height)
		 (setf (aref captcha (+ i di) j)
		       (aref letter-images (+ letter-offset di) j)))))

	;; squish and save the captcha
	(with-open-file (s output-path
			 :direction :output
			 :if-exists :supersede)
	  (write-pgm (tilt-a-whirl captcha
				   :whirl-angle (* (if (zerop (random 2))
						       -1
						       1)
						   (+ 60 (random 30)))
				   :radius-scale (+ 1.0 (random 0.8)))
		     s)))))


(defun tilt-a-whirl (captcha &key (whirl-angle 90) (pinch 0) (radius-scale 1.0))
  "Twist up a picture like the GIMP's `Whirl and Pinch' plug-in by
Federico Mena Quintero and Scott Goehring.  The bugs here are my
fault, of course."
  (let* ((width (car (array-dimensions captcha)))
	 (height (cadr (array-dimensions captcha)))
	 (new-captcha (make-array (list width height) :element-type 'fixnum))
	 (center-x (/ (- width 1)  2.0))
	 (center-y (/ (- height 1) 2.0))
	 (scale-x (if (< width height) (/ height width)  1.0))
	 (scale-y (if (>= width height) (/ width height) 1.0 ))
	 (whirl (* whirl-angle (/ pi 180)))
	 (radius (max (/ width 2.0) (/ height 2.0)))
	 (radius2 (* radius radius radius-scale)))
    (flet ((calc-coords (x y)
	     ;; distances to center, scaled
	     (let* ((dx (* (- x center-x) scale-x))
		    (dy (* (- y center-y) scale-y))
		    ;; distance^2 to center of ellipse
		    (d (+ (* dx dx) (* dy dy)))
		    (inside (< d radius2)))
	       (if inside
		   (let ((dist (/ (sqrt (/ d radius-scale)) radius)))
		     ;; pinch
		     (let* ((factor (expt (sin (* (/ pi 2) dist)) (- pinch)))
			    (dx (* dx factor))
			    (dy (* dy factor)))
		       ;; whirl
		       (let* ((factor (- 1.0 dist))
			      (ang (* whirl factor factor))
			      (sina (sin ang))
			      (cosa (cos ang)))
			 (values
			  ;; x
			  (+ (/ (- (* cosa dx) (* sina dy))
				scale-x)
			     center-x)
			  ;; y
			  (+ (/ (+ (* sina dx) (* cosa dy))
				scale-y)
			     center-y)
			  ;; inside
			  t))))
		   ;; not inside
		   (values x y nil)))))

      (dotimes (row  (round (/ height 2)))
	(let ((top-index 0)
	      (bottom-index (- width 1)))
	  (dotimes (col width)
	    (multiple-value-bind (cx cy inside)
		(calc-coords col row)
	      (if inside
		  (progn
		    ;; top
		    (let ((ix (if (minusp cx) (ceiling cx) (floor cx)))
			  (iy (if (minusp cy) (ceiling cy) (floor cy))))
		      (setf (aref new-captcha top-index row)
			    (bilerp (/ cx center-x) (/ cy center-y)
				    (get-pixel captcha ix iy)
				    (get-pixel captcha (1+ ix) iy)
				    (get-pixel captcha ix (1+ iy))
				    (get-pixel captcha (1+ ix) (1+ iy))))
		      (incf top-index))
		    ;; bottom
		    (let ((cx (+ center-x (- center-x cx)))
			  (cy (+ center-y (- center-y cy))))
		      (let ((ix (if (minusp cx) (ceiling cx) (floor cx)))
			    (iy (if (minusp cy) (ceiling cy) (floor cy))))
			(setf (aref new-captcha bottom-index (- height row 1))
			      (bilerp (/ cx center-x) (/ cy center-y)
				      (get-pixel captcha ix iy)
				      (get-pixel captcha (1+ ix) iy)
				      (get-pixel captcha ix (1+ iy))
				      (get-pixel captcha (1+ ix) (1+ iy))))
			(decf bottom-index))))
		  ;; not inside, just copy
		  (progn
		    (setf (aref new-captcha top-index row)
			  (aref captcha top-index row))
		    (incf top-index)
		    (setf (aref new-captcha bottom-index (- height row 1))
			  (aref captcha bottom-index (- height row 1)))
		    (decf bottom-index)
		    ))))))
      new-captcha
      )))
		      
			
(defun get-pixel (pixels x y)
  (let ((width (car (array-dimensions pixels)))
	(height (cadr (array-dimensions pixels))))
    (if (or (>= x width)
	    (>= y height)
	    (< x 0)
	    (< y 0))
	;; background color
	255
	(aref pixels x y))))

	

(defun bilerp (x y l m f g)
  "Bilinear interpolation, I imagine."
  (let ((x (rem x 1.0))
	(y (rem y 1.0)))
    (when (minusp x) (incf x 1.0))
    (when (minusp y) (incf y 1.0))
    (let ((m0 (+ (* l (- 1.0 x)) (* m x)))
	  (m1 (+ (* f (- 1.0 x)) (* g x))))
      (let ((retval (round (+ (* (- 1.0 y) m0) (* y m1)))))
	retval))))


;;;; Our Portable Graymap library
;;;; PGM is a dirt simple format for images.
      
(defun write-pgm (bytes &optional (stream *standard-output*))
  "Write a Plain PGM file to STREAM using the dimensions and grey
values in BYTES."
  (let* ((dimensions (array-dimensions bytes))
	 (width (car dimensions))
	 (height (cadr dimensions)))
    ;; Plain PGM header: magic number, width, and height, maximum grey
    ;; value.
    (format stream "P2~%~A ~A~%256~%"  width height)
    ;; Each raster is HEIGHT rows long
    (let ((current-line-length 0)
	  (*print-pretty* nil))
      (dotimes (i height)
	(dotimes (j width)
	  (princ (aref bytes j i) stream)
	  (write-char #\space stream)
	  ;; No line in the output file may be longer than 70 characters
	  ;; We assume a maximum of 4 characters per pixel
	  (unless (< (incf current-line-length 4) 70)
	    (terpri stream)
	    (setf current-line-length 0)))))
    (terpri stream)))


(defun read-pgm (&optional (stream *standard-input*))
  "Read a Plain PGM file from STREAM, return a rank-2 array ."
  (let ((*read-eval* nil)
	(*package* (or (find-package "PGM-READ")
		       (make-package "PGM-READ"))))
    (let ((magic-number (read stream)))
      (assert (string= (symbol-name magic-number) "P2")
	      nil
	      "Invalid magic number for PGM file.  Expected P2, got ~A."
	      magic-number)
      (let ((width (read stream))
	    (height (read stream))
	    (maximum-grey-value (read stream)))
	;; don't want to cons perversely
	(assert (and (typep width '(integer 1 2048))
		     (typep height '(integer 1 2048))
		     (typep maximum-grey-value '(integer 1 #.(expt 2 24))))
		nil
		"PGM file is entirely too special.  ~
                 Width ~A, height ~A, maximum grey value ~A."
		width height maximum-grey-value)
	(let ((pgm-data (make-array (list width height)
				    :element-type 'fixnum)))
	  (dotimes (i height)
	    (dotimes (j width)
	      (setf (aref pgm-data j i) (read stream))))
	  pgm-data)))))


(defun pgm-test ()  
  "Perform a basic test of reading and writings PGM files."
  (let ((bits (make-array '(1024 768))))
    (dotimes (i 1024)
      (dotimes (j 768)
	;; we expect horizontal stripes with vertical gradients in a
	;; 1024x768 image
	(setf (aref bits i j)
	      (rem (* j 8) 256))))
    (with-open-file (s "/tmp/test.pgm" :direction :output
		     :if-exists :supersede)
      (write-pgm bits s))
    ;; now let's read it back
    (with-open-file (s "/tmp/test.pgm")
      (let ((echo-bits (read-pgm s)))
	(dotimes (i 1024)
	  (dotimes (j 768)
	    (assert (= (aref bits i j)
		       (aref echo-bits i j))
		    nil "Test failed: The PGM was corrupted at (~A,~A)." i j))))))
  "Test passed!")

